// The MIT License (MIT)
//
// Copyright (c) 2017-2020 Uber Technologies Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package invariants

import (
	"testing"

	"github.com/golang/mock/gomock"
	"github.com/stretchr/testify/require"
	"github.com/stretchr/testify/suite"

	"github.com/uber/cadence/service/worker/scanner/executions/common"
)

type InvariantManagerSuite struct {
	*require.Assertions
	suite.Suite
	controller *gomock.Controller
}

func TestInvariantManagerSuite(t *testing.T) {
	suite.Run(t, new(InvariantManagerSuite))
}

func (s *InvariantManagerSuite) SetupTest() {
	s.Assertions = require.New(s.T())
	s.controller = gomock.NewController(s.T())
}

func (s *InvariantManagerSuite) TearDownTest() {
	s.controller.Finish()
}

func (s *InvariantManagerSuite) TestRunChecks() {
	testCases := []struct {
		checkResults []common.CheckResult
		expected     common.ManagerCheckResult
	}{
		{
			checkResults: nil,
			expected: common.ManagerCheckResult{
				CheckResultType: common.CheckResultTypeHealthy,
				CheckResults:    nil,
			},
		},
		{
			checkResults: []common.CheckResult{
				{
					CheckResultType: common.CheckResultTypeHealthy,
					InvariantType:   common.InvariantType("first"),
					Info:            "invariant 1 info",
					InfoDetails:     "invariant 1 info details",
				},
				{
					CheckResultType: common.CheckResultTypeFailed,
					InvariantType:   common.InvariantType("second"),
					Info:            "invariant 2 info",
					InfoDetails:     "invariant 2 info details",
				},
			},
			expected: common.ManagerCheckResult{
				CheckResultType:          common.CheckResultTypeFailed,
				DeterminingInvariantType: common.InvariantTypePtr("second"),
				CheckResults: []common.CheckResult{
					{
						CheckResultType: common.CheckResultTypeHealthy,
						InvariantType:   common.InvariantType("first"),
						Info:            "invariant 1 info",
						InfoDetails:     "invariant 1 info details",
					},
					{
						CheckResultType: common.CheckResultTypeFailed,
						InvariantType:   common.InvariantType("second"),
						Info:            "invariant 2 info",
						InfoDetails:     "invariant 2 info details",
					},
				},
			},
		},
		{
			checkResults: []common.CheckResult{
				{
					CheckResultType: common.CheckResultTypeHealthy,
					InvariantType:   common.InvariantType("first"),
					Info:            "invariant 1 info",
					InfoDetails:     "invariant 1 info details",
				},
				{
					CheckResultType: common.CheckResultTypeCorrupted,
					InvariantType:   common.InvariantType("second"),
					Info:            "invariant 2 info",
					InfoDetails:     "invariant 2 info details",
				},
			},
			expected: common.ManagerCheckResult{
				CheckResultType:          common.CheckResultTypeCorrupted,
				DeterminingInvariantType: common.InvariantTypePtr("second"),
				CheckResults: []common.CheckResult{
					{
						CheckResultType: common.CheckResultTypeHealthy,
						InvariantType:   common.InvariantType("first"),
						Info:            "invariant 1 info",
						InfoDetails:     "invariant 1 info details",
					},
					{
						CheckResultType: common.CheckResultTypeCorrupted,
						InvariantType:   common.InvariantType("second"),
						Info:            "invariant 2 info",
						InfoDetails:     "invariant 2 info details",
					},
				},
			},
		},
		{
			checkResults: []common.CheckResult{
				{
					CheckResultType: common.CheckResultTypeHealthy,
					InvariantType:   common.InvariantType("first"),
					Info:            "invariant 1 info",
					InfoDetails:     "invariant 1 info details",
				},
				{
					CheckResultType: common.CheckResultTypeHealthy,
					InvariantType:   common.InvariantType("second"),
					Info:            "invariant 2 info",
					InfoDetails:     "invariant 2 info details",
				},
			},
			expected: common.ManagerCheckResult{
				CheckResultType:          common.CheckResultTypeHealthy,
				DeterminingInvariantType: nil,
				CheckResults: []common.CheckResult{
					{
						CheckResultType: common.CheckResultTypeHealthy,
						InvariantType:   common.InvariantType("first"),
						Info:            "invariant 1 info",
						InfoDetails:     "invariant 1 info details",
					},
					{
						CheckResultType: common.CheckResultTypeHealthy,
						InvariantType:   common.InvariantType("second"),
						Info:            "invariant 2 info",
						InfoDetails:     "invariant 2 info details",
					},
				},
			},
		},
		{
			checkResults: []common.CheckResult{
				{
					CheckResultType: common.CheckResultTypeHealthy,
					InvariantType:   common.InvariantType("first"),
					Info:            "invariant 1 info",
					InfoDetails:     "invariant 1 info details",
				},
				{
					CheckResultType: common.CheckResultTypeCorrupted,
					InvariantType:   common.InvariantType("second"),
					Info:            "invariant 2 info",
					InfoDetails:     "invariant 2 info details",
				},
				{
					CheckResultType: common.CheckResultTypeFailed,
					InvariantType:   common.InvariantType("third"),
					Info:            "invariant 3 info",
					InfoDetails:     "invariant 3 info details",
				},
				{
					CheckResultType: common.CheckResultTypeHealthy,
					InvariantType:   common.InvariantType("forth"),
					Info:            "invariant 4 info",
					InfoDetails:     "invariant 4 info details",
				},
			},
			expected: common.ManagerCheckResult{
				CheckResultType:          common.CheckResultTypeFailed,
				DeterminingInvariantType: common.InvariantTypePtr("third"),
				CheckResults: []common.CheckResult{
					{
						CheckResultType: common.CheckResultTypeHealthy,
						InvariantType:   common.InvariantType("first"),
						Info:            "invariant 1 info",
						InfoDetails:     "invariant 1 info details",
					},
					{
						CheckResultType: common.CheckResultTypeCorrupted,
						InvariantType:   common.InvariantType("second"),
						Info:            "invariant 2 info",
						InfoDetails:     "invariant 2 info details",
					},
					{
						CheckResultType: common.CheckResultTypeFailed,
						InvariantType:   common.InvariantType("third"),
						Info:            "invariant 3 info",
						InfoDetails:     "invariant 3 info details",
					},
					{
						CheckResultType: common.CheckResultTypeHealthy,
						InvariantType:   common.InvariantType("forth"),
						Info:            "invariant 4 info",
						InfoDetails:     "invariant 4 info details",
					},
				},
			},
		},
	}

	for _, tc := range testCases {
		invariants := make([]common.Invariant, len(tc.checkResults), len(tc.checkResults))
		for i := 0; i < len(tc.checkResults); i++ {
			mockInvariant := common.NewMockInvariant(s.controller)
			mockInvariant.EXPECT().Check(gomock.Any()).Return(tc.checkResults[i])
			invariants[i] = mockInvariant
		}
		manager := &invariantManager{
			invariants: invariants,
		}
		s.Equal(tc.expected, manager.RunChecks(common.Execution{}))
	}
}

func (s *InvariantManagerSuite) TestRunFixes() {
	testCases := []struct {
		fixResults []common.FixResult
		expected   common.ManagerFixResult
	}{
		{
			fixResults: nil,
			expected: common.ManagerFixResult{
				FixResultType:            common.FixResultTypeSkipped,
				DeterminingInvariantType: nil,
				FixResults:               nil,
			},
		},
		{
			fixResults: []common.FixResult{
				{
					FixResultType: common.FixResultTypeFixed,
					InvariantType: common.InvariantType("first"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 1 check info",
						InfoDetails:     "invariant 1 check info details",
					},
					Info:        "invariant 1 info",
					InfoDetails: "invariant 1 info details",
				},
				{
					FixResultType: common.FixResultTypeFailed,
					InvariantType: common.InvariantType("second"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 2 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 2 info",
					InfoDetails: "invariant 2 info details",
				},
			},
			expected: common.ManagerFixResult{
				FixResultType:            common.FixResultTypeFailed,
				DeterminingInvariantType: common.InvariantTypePtr("second"),
				FixResults: []common.FixResult{
					{
						FixResultType: common.FixResultTypeFixed,
						InvariantType: common.InvariantType("first"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 1 check info",
							InfoDetails:     "invariant 1 check info details",
						},
						Info:        "invariant 1 info",
						InfoDetails: "invariant 1 info details",
					},
					{
						FixResultType: common.FixResultTypeFailed,
						InvariantType: common.InvariantType("second"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 2 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 2 info",
						InfoDetails: "invariant 2 info details",
					},
				},
			},
		},
		{
			fixResults: []common.FixResult{
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("first"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 1 check info",
						InfoDetails:     "invariant 1 check info details",
					},
					Info:        "invariant 1 info",
					InfoDetails: "invariant 1 info details",
				},
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("second"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 2 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 2 info",
					InfoDetails: "invariant 2 info details",
				},
			},
			expected: common.ManagerFixResult{
				FixResultType:            common.FixResultTypeSkipped,
				DeterminingInvariantType: nil,
				FixResults: []common.FixResult{
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("first"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 1 check info",
							InfoDetails:     "invariant 1 check info details",
						},
						Info:        "invariant 1 info",
						InfoDetails: "invariant 1 info details",
					},
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("second"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 2 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 2 info",
						InfoDetails: "invariant 2 info details",
					},
				},
			},
		},
		{
			fixResults: []common.FixResult{
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("first"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 1 check info",
						InfoDetails:     "invariant 1 check info details",
					},
					Info:        "invariant 1 info",
					InfoDetails: "invariant 1 info details",
				},
				{
					FixResultType: common.FixResultTypeFixed,
					InvariantType: common.InvariantType("second"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 2 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 2 info",
					InfoDetails: "invariant 2 info details",
				},
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("third"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 3 check info",
						InfoDetails:     "invariant 3 check info details",
					},
					Info:        "invariant 3 info",
					InfoDetails: "invariant 3 info details",
				},
			},
			expected: common.ManagerFixResult{
				FixResultType:            common.FixResultTypeFixed,
				DeterminingInvariantType: common.InvariantTypePtr("second"),
				FixResults: []common.FixResult{
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("first"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 1 check info",
							InfoDetails:     "invariant 1 check info details",
						},
						Info:        "invariant 1 info",
						InfoDetails: "invariant 1 info details",
					},
					{
						FixResultType: common.FixResultTypeFixed,
						InvariantType: common.InvariantType("second"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 2 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 2 info",
						InfoDetails: "invariant 2 info details",
					},
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("third"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 3 check info",
							InfoDetails:     "invariant 3 check info details",
						},
						Info:        "invariant 3 info",
						InfoDetails: "invariant 3 info details",
					},
				},
			},
		},
		{
			fixResults: []common.FixResult{
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("first"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 1 check info",
						InfoDetails:     "invariant 1 check info details",
					},
					Info:        "invariant 1 info",
					InfoDetails: "invariant 1 info details",
				},
				{
					FixResultType: common.FixResultTypeFixed,
					InvariantType: common.InvariantType("second"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 2 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 2 info",
					InfoDetails: "invariant 2 info details",
				},
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("third"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 3 check info",
						InfoDetails:     "invariant 3 check info details",
					},
					Info:        "invariant 3 info",
					InfoDetails: "invariant 3 info details",
				},
				{
					FixResultType: common.FixResultTypeFailed,
					InvariantType: common.InvariantType("forth"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 4 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 4 info",
					InfoDetails: "invariant 4 info details",
				},
			},
			expected: common.ManagerFixResult{
				FixResultType:            common.FixResultTypeFailed,
				DeterminingInvariantType: common.InvariantTypePtr("forth"),
				FixResults: []common.FixResult{
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("first"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 1 check info",
							InfoDetails:     "invariant 1 check info details",
						},
						Info:        "invariant 1 info",
						InfoDetails: "invariant 1 info details",
					},
					{
						FixResultType: common.FixResultTypeFixed,
						InvariantType: common.InvariantType("second"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 2 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 2 info",
						InfoDetails: "invariant 2 info details",
					},
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("third"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 3 check info",
							InfoDetails:     "invariant 3 check info details",
						},
						Info:        "invariant 3 info",
						InfoDetails: "invariant 3 info details",
					},
					{
						FixResultType: common.FixResultTypeFailed,
						InvariantType: common.InvariantType("forth"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 4 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 4 info",
						InfoDetails: "invariant 4 info details",
					},
				},
			},
		},
		{
			fixResults: []common.FixResult{
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("first"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 1 check info",
						InfoDetails:     "invariant 1 check info details",
					},
					Info:        "invariant 1 info",
					InfoDetails: "invariant 1 info details",
				},
				{
					FixResultType: common.FixResultTypeFailed,
					InvariantType: common.InvariantType("second"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 4 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 4 info",
					InfoDetails: "invariant 4 info details",
				},
				{
					FixResultType: common.FixResultTypeFixed,
					InvariantType: common.InvariantType("third"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeCorrupted,
						Info:            "invariant 2 check info",
						InfoDetails:     "invariant 2 check info details",
					},
					Info:        "invariant 2 info",
					InfoDetails: "invariant 2 info details",
				},
				{
					FixResultType: common.FixResultTypeSkipped,
					InvariantType: common.InvariantType("forth"),
					CheckResult: common.CheckResult{
						CheckResultType: common.CheckResultTypeHealthy,
						Info:            "invariant 3 check info",
						InfoDetails:     "invariant 3 check info details",
					},
					Info:        "invariant 3 info",
					InfoDetails: "invariant 3 info details",
				},
			},
			expected: common.ManagerFixResult{
				FixResultType:            common.FixResultTypeFailed,
				DeterminingInvariantType: common.InvariantTypePtr("second"),
				FixResults: []common.FixResult{
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("first"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 1 check info",
							InfoDetails:     "invariant 1 check info details",
						},
						Info:        "invariant 1 info",
						InfoDetails: "invariant 1 info details",
					},
					{
						FixResultType: common.FixResultTypeFailed,
						InvariantType: common.InvariantType("second"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 4 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 4 info",
						InfoDetails: "invariant 4 info details",
					},
					{
						FixResultType: common.FixResultTypeFixed,
						InvariantType: common.InvariantType("third"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeCorrupted,
							Info:            "invariant 2 check info",
							InfoDetails:     "invariant 2 check info details",
						},
						Info:        "invariant 2 info",
						InfoDetails: "invariant 2 info details",
					},
					{
						FixResultType: common.FixResultTypeSkipped,
						InvariantType: common.InvariantType("forth"),
						CheckResult: common.CheckResult{
							CheckResultType: common.CheckResultTypeHealthy,
							Info:            "invariant 3 check info",
							InfoDetails:     "invariant 3 check info details",
						},
						Info:        "invariant 3 info",
						InfoDetails: "invariant 3 info details",
					},
				},
			},
		},
	}

	for _, tc := range testCases {
		invariants := make([]common.Invariant, len(tc.fixResults), len(tc.fixResults))
		for i := 0; i < len(tc.fixResults); i++ {
			mockInvariant := common.NewMockInvariant(s.controller)
			mockInvariant.EXPECT().Fix(gomock.Any()).Return(tc.fixResults[i])
			invariants[i] = mockInvariant
		}
		manager := &invariantManager{
			invariants: invariants,
		}
		s.Equal(tc.expected, manager.RunFixes(common.Execution{}))
	}
}
